@@ -1,9 +1,9 @@
 /* Test of case-insensitive string comparison function.
-   Copyright (C) 2007-2013 Free Software Foundation, Inc.
+   Copyright (C) 2007-2025 Free Software Foundation, Inc.
 
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
-   the Free Software Foundation; either version 3 of the License, or
+   the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.
 
    This program is distributed in the hope that it will be useful,
@@ -12,7 +12,7 @@
    GNU General Public License for more details.
 
    You should have received a copy of the GNU General Public License
-   along with this program.  If not, see <http://www.gnu.org/licenses/>.  */
+   along with this program.  If not, see <https://www.gnu.org/licenses/>.  */
 
 /* Written by Bruno Haible <bruno@clisp.org>, 2007.  */
 
@@ -51,5 +51,5 @@ main ()
   ASSERT (mbscasecmp ("turkish", "TURK\304\260SH") == 0);
   ASSERT (mbscasecmp ("TURK\304\260SH", "turkish") == 0);
 
-  return 0;
+  return test_exit_status;
 }
